HAVE_VTK = False
try:
    import vtk
    HAVE_VTK = True
    from vtk.util.numpy_support import vtk_to_numpy
except ImportError:
    pass

from ngff_zarr import ngff_image, to_ngff_image


def vtk_image_to_ngff_image(image):
    array = vtk_to_numpy(image.GetPointData().GetScalars())
    dimensions = list(image.GetDimensions())
    array.shape = dimensions[::-1]

    origin = image.GetOrigin()
    translation = { 'x': origin[0], 'y': origin[1], 'z': origin[2] }

    spacing = image.GetSpacing()
    scale = { 'x': spacing[0], 'y': spacing[1], 'z': spacing[2] }

    ngff_image = to_ngff_image(array, scale=scale, translation=translation)

    return ngff_image

def vtk_polydata_to_vtkjs(point_set):
    array = vtk_to_numpy(point_set.GetPoints().GetData())
    return array



import asyncio
import os
import queue
import threading
from imjoy_rpc import api
from typing import List
from IPython.display import display, HTML
from IPython.lib import backgroundjobs as bg
import uuid

from ._type_aliases import Gaussians, Style, Image, PointSet
from ._initialization_params import init_params_dict
from ._method_types import deferred_methods
from .imjoy import register_itkwasm_imjoy_codecs
from .integrations import _detect_render_type, _get_viewer_image, _get_viewer_point_set
from .integrations.environment import ENVIRONMENT, Env
from .render_types import RenderType
from .viewer_config import ITK_VIEWER_SRC, PYDATA_SPHINX_HREF, MUI_HREF
from imjoy_rpc import register_default_codecs

__all__ = [
    "Viewer",
    "view",
]

_viewer_count = 1
_codecs_registered = False


class ViewerRPC:
    """Viewer remote procedure interface."""

    def __init__(
        self, ui_collapsed=True, rotate=False, ui="pydata-sphinx", data=None
    ):
        global _codecs_registered
        """Create a viewer."""
        # Register codecs if they haven't been already
        if not _codecs_registered:
            register_default_codecs()
            register_itkwasm_imjoy_codecs()
            _codecs_registered = True

        self._init_viewer_kwargs = dict(ui_collapsed=ui_collapsed, rotate=rotate, ui=ui)
        self.init_data = data
        self.img = display(HTML(f'<div />'), display_id=str(uuid.uuid4()))
        self.wid = None
        if ENVIRONMENT is not Env.JUPYTERLITE:
            self.viewer_event = threading.Event()
            self.data_event = threading.Event()

    async def setup(self):
        pass

    async def run(self, ctx):
        """ImJoy plugin setup function."""
        global _viewer_count
        ui = self._init_viewer_kwargs.get("ui", None)
        if ui == "pydata-sphinx":
            config = {
                "uiMachineOptions": {
                    "href": PYDATA_SPHINX_HREF,
                    "export": "default",
                }
            }
        elif ui == "mui":
            config = {
                "uiMachineOptions": {
                    "href": MUI_HREF,
                    "export": "default",
                }
            }
        elif ui != "reference":
            config = ui
        else:
            config = {}
        config['maxConcurrency'] = os.cpu_count() * 2

        itk_viewer = await api.createWindow(
            name=f"itkwidgets viewer {_viewer_count}",
            type="itk-vtk-viewer",
            src=ITK_VIEWER_SRC,
            fullscreen=True,
            data=self.init_data,
            # config should be a python data dictionary and can't be a string e.g. 'pydata-sphinx',
            config=config,
        )
        _viewer_count += 1
        if ENVIRONMENT is not Env.JUPYTERLITE:
            itk_viewer.registerEventListener(
                'renderedImageAssigned', self.set_event
            )
            # Once the viewer has been created any queued requests can be run
            asyncio.get_running_loop().call_soon_threadsafe(self.viewer_event.set)

        self.set_default_ui_values(itk_viewer)
        self.itk_viewer = itk_viewer
        self.wid = self.itk_viewer.config.window_id

        # Create the initial screenshot
        await self.create_screenshot()
        # Wait and then update the screenshot in case rendered level changed
        await asyncio.sleep(10)
        await self.create_screenshot()
        # Set up an event listener so that the embedded
        # screenshot is updated when the user requests
        itk_viewer.registerEventListener(
            'screenshotTaken', self.update_screenshot
        )

    def set_default_ui_values(self, itk_viewer):
        settings = init_params_dict(itk_viewer)
        for key, value in self._init_viewer_kwargs.items():
            if key in settings.keys():
                settings[key](value)

    async def create_screenshot(self):
        base64_image = await self.itk_viewer.captureImage()
        self.update_screenshot(base64_image)

    def update_screenshot(self, base64_image):
        html = HTML(
            f'''
                <img id="screenshot_{self.wid}" src={base64_image}>
                <script type="text/javascript">
                    var image = document.getElementById("screenshot_{self.wid}");
                    image.src = "{base64_image}";
                    var viewer = document.getElementById("{self.wid}");
                    // Hide the static image if the Viewer is visible
                    image.style.display = viewer ? "none" : "block";
                </script>
            ''')
        self.img.display(html)

    def set_event(self, event_data):
        # Once the data has been set the deferred queue requests can be run
        asyncio.get_running_loop().call_soon_threadsafe(self.data_event.set)


class Viewer:
    """Pythonic Viewer class."""

    def __init__(
        self, ui_collapsed=True, rotate=False, ui="pydata-sphinx", **add_data_kwargs
    ):
        input_data = self.input_data(add_data_kwargs)
        data = self.init_data(input_data)
        """Create a viewer."""
        self.viewer_rpc = ViewerRPC(
            ui_collapsed=ui_collapsed, rotate=rotate, ui=ui, data=data
        )
        if ENVIRONMENT is not Env.JUPYTERLITE:
            self.bg_jobs = bg.BackgroundJobManager()
            self.queue = queue.Queue()
            self.deferred_queue = queue.Queue()
            self.bg_thread = self.bg_jobs.new(self.queue_worker)
        api.export(self.viewer_rpc)

    @property
    def loop(self):
        return asyncio.get_running_loop()

    @property
    def has_viewer(self):
        return hasattr(self.viewer_rpc, 'itk_viewer')

    def input_data(self, init_data_kwargs):
        input_options = ["data", "image", "label_image", "point_set"]
        inputs = []
        for option in input_options:
            data = init_data_kwargs.get(option, None)
            if data is not None:
                inputs.append((option, data))
        return inputs

    def init_data(self, input_data):
        _init_data = {}
        result= None
        for (input_type, data) in input_data:
            render_type = _detect_render_type(data, input_type)
            if render_type is RenderType.IMAGE:
                if input_type == 'label_image':
                    result = _get_viewer_image(data, label=True)
                    render_type = RenderType.LABELIMAGE
                else:
                    result = _get_viewer_image(data, label=False)
            elif render_type is RenderType.POINT_SET:
                result = _get_viewer_point_set(data)
            if result is None:
                raise RuntimeError(f"Could not process the viewer {input_type}")
            _init_data[render_type.value] = result
        return _init_data

    async def run_queued_requests(self):
        def _run_queued_requests(queue):
            method_name, args, kwargs = queue.get()
            fn = getattr(self.viewer_rpc.itk_viewer, method_name)
            self.loop.call_soon_threadsafe(asyncio.ensure_future, fn(*args, **kwargs))

        # Wait for the viewer to be created
        self.viewer_rpc.viewer_event.wait()
        while self.queue.qsize():
            _run_queued_requests(self.queue)
        # Wait for the data to be set
        self.viewer_rpc.data_event.wait()
        while self.deferred_queue.qsize():
            _run_queued_requests(self.deferred_queue)

    def queue_worker(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        task = loop.create_task(self.run_queued_requests())
        loop.run_until_complete(task)

    def queue_request(self, method, *args, **kwargs):
        if ENVIRONMENT is Env.JUPYTERLITE or self.has_viewer:
            fn = getattr(self.viewer_rpc.itk_viewer, method)
            fn(*args, **kwargs)
        elif method in deferred_methods():
            self.deferred_queue.put((method, args, kwargs))
        else:
            self.queue.put((method, args, kwargs))

    def set_annotations_enabled(self, enabled: bool):
        self.queue_request('setAnnotationsEnabled', enabled)

    def set_axes_enabled(self, enabled: bool):
        self.queue_request('setAxesEnabled', enabled)

    def set_background_color(self, bgColor: List[float]):
        self.queue_request('setBackgroundColor', bgColor)

    def set_image(self, image: Image):
        render_type = _detect_render_type(image, 'image')
        if render_type is RenderType.IMAGE:
            image = _get_viewer_image(image, label=False)
            self.queue_request('setImage', image)
        elif render_type is RenderType.POINT_SET:
            image = _get_viewer_point_set(image)
            self.queue_request('setPointSets', image)

    def set_image_blend_mode(self, mode: str):
        self.queue_request('setImageBlendMode', mode)

    def set_image_color_map(self, colorMap: str):
        self.queue_request('setImageColorMap', colorMap)

    def set_image_color_range(self, range: List[float]):
        self.queue_request('setImageColorRange', range)

    def set_image_color_range_bounds(self, range: List[float]):
        self.queue_request('setImageColorRangeBounds', range)

    def set_image_component_visibility(self, visibility: bool):
        self.queue_request('setImageComponentVisibility', visibility)

    def set_image_gradient_opacity(self, opacity: float):
        self.queue_request('setImageGradientOpacity', opacity)

    def set_image_gradient_opacity_scale(self, min: float):
        self.queue_request('setImageGradientOpacityScale', min)

    def set_image_interpolation_enabled(self, enabled: bool):
        self.queue_request('setImageInterpolationEnabled', enabled)

    def set_image_piecewise_function_gaussians(self, gaussians: Gaussians):
        self.queue_request('setImagePiecewiseFunctionGaussians', gaussians)

    def set_image_shadow_enabled(self, enabled: bool):
        self.queue_request('setImageShadowEnabled', enabled)

    def set_image_volume_sample_distance(self, distance: float):
        self.queue_request('setImageVolumeSampleDistance', distance)

    def set_image_volume_scattering_blend(self, scattering_blend: float):
        self.queue_request('setImageVolumeScatteringBlend', scattering_blend)

    def set_label_image(self, label_image: Image):
        render_type = _detect_render_type(label_image, 'image')
        if render_type is RenderType.IMAGE:
            label_image = _get_viewer_image(label_image, label=True)
            self.queue_request('setLabelImage', label_image)
        elif render_type is RenderType.POINT_SET:
            label_image = _get_viewer_point_set(label_image)
            self.queue_request('setPointSets', label_image)

    def set_label_image_blend(self, blend: float):
        self.queue_request('setLabelImageBlend', blend)

    def set_label_image_label_names(self, names: List[str]):
        self.queue_request('setLabelImageLabelNames', names)

    def set_label_image_lookup_table(self, lookupTable: str):
        self.queue_request('setLabelImageLookupTable', lookupTable)

    def set_label_image_weights(self, weights: float):
        self.queue_request('setLabelImageWeights', weights)

    def select_layer(self, name: str):
        self.queue_request('selectLayer', name)

    def set_layer_visibility(self, visible: bool):
        self.queue_request('setLayerVisibility', visible)

    def add_point_set(self, pointSet: PointSet):
        pointSet = _get_viewer_point_set(pointSet)
        self.queue_request('addPointSet', pointSet)
    def set_point_set(self, pointSet: PointSet):
        pointSet = _get_viewer_point_set(pointSet)
        self.queue_request('setPointSets', pointSet)

    def set_rendering_view_container_style(self, containerStyle: Style):
        self.queue_request('setRenderingViewContainerStyle', containerStyle)

    def set_rotate(self, enabled: bool):
        self.queue_request('setRotateEnabled', enabled)

    def set_ui_collapsed(self, collapsed: bool):
        self.queue_request('setUICollapsed', collapsed)

    def set_units(self, units: str):
        self.queue_request('setUnits', units)

    def set_view_mode(self, mode: str):
        self.queue_request('setViewMode', mode)

    def set_x_slice(self, position: float):
        self.queue_request('setXSlice', position)

    def set_y_slice(self, position: float):
        self.queue_request('setYSlice', position)

    def set_z_slice(self, position: float):
        self.queue_request('setZSlice', position)


def view(data=None, **kwargs):
    """View the image and/or point set.

    Creates and returns an ImJoy plugin ipywidget to visualize an image, and/or
    point set.

    The image can be 2D or 3D. The type of the image can be an numpy.array,
    itk.Image, or vtk.vtkImageData.

    A point set can be visualized. The type of the point set can be an
    numpy.array (Nx3 array of point positions).

    Parameters
    ----------
    General Interface
    ^^^^^^^^^^^^^^^^^
    ui_collapsed : bool, default: True
        Collapse the native widget user interface.
    rotate : bool, default: False
        Continuously rotate the camera around the scene in volume rendering
        mode.
    annotations: bool, default: True
        Display annotations describing orientation and the value of a
        mouse-position-based data probe.
    axes: bool, default: False
        Display axes.
    bg_color: (red, green, blue) tuple, components from 0.0 to 1.0
        Background color. Default is based on the current Jupyter theme.
    container_style: dict
        The CSS style for the rendering view `div`'s.

    Images
    ^^^^^^
    image : array_like, itk.Image, or vtk.vtkImageData
        The 2D or 3D image to visualize.
    label_image: array_like, itk.Image, or vtk.vtkImageData
        The 2D or 3D label map to visualize. If an image is also provided, the
        label map must have the same size.
    label_blend: float, default: 0.5
        Label map blend with intensity image, from 0.0 to 1.0.
    label_names: list of (label_value, label_name)
        String names associated with the integer label values.
    label_lut: string, default: 'glasbey'
        Lookup table for the label map.
    label_weights: float
        The rendering weight assigned to current label. Values range from 0.0
        to 1.0.
    color_range: list, default: The [min, max] range of the data values
        The [min, max] range of intensity values mapped to colors for the given
        image component identified by name.
    color_bounds: list, default: The [min, max] range of the data values
        The [min, max] range of intensity values for color maps that provide a
        bounds for user inputs.
    cmap: string, default: 'Grayscale'
        The color map for the current component/channel.
    x_slice: float, default: None
        The position in world space of the X slicing plane.
    y_slice: float, default: None
        The position in world space of the Y slicing plane.
    z_slice: float, default: None
        The position in world space of the Z slicing plane.
    interpolation: bool, deafult: True
        Linear as opposed to nearest neighbor interpolation for image slices.
        Note: Interpolation is not currently supported with label maps.
    gradient_opacity: float, default: 0.5
        Gradient opacity for composite volume rendering, in the range
        (0.0, 1.0].
    gradient_opacity_scale: float, default: 0.5
        Gradient opacity scale for composite volume rendering, in the range
        (0.0, 1.0].
    gaussians: dict
        Volume rendering opacity transfer function Gaussian parameters. For
        each image component, multiple Gaussians can be specified.
        Default Gaussian parameters:
          {'position': 0.5, 'height': 1, 'width': 0.5, 'xBias': 0.51, 'yBias': 0.4}
    blend_mode: string, default: 'Composite'
        Volume rendering blend mode. Supported modes: 'Composite', 'Maximum',
        'Minimum', 'Average'.
    component_visible: bool, default: True
        The given image intensity component index's visibility.
    shadow_enabled: bool, default: True
        Whether to used gradient-based shadows in the volume rendering.
    view_mode: 'XPlane', 'YPlane', 'ZPlane', or 'Volume', default: 'Volume'
        Only relevant for 3D scenes.
        Viewing mode:
            'XPlane': x-plane
            'YPlane': y-plane
            'ZPlane': z-plane
            'Volume': volume rendering
    layer: string
        Select the layer identified by `name` in the user interface.
    layer_visible: bool, deafult: True
        Whether the current layer is visible.

    Point Set
    ^^^^^^^^^^
    point_set: point set
        The point set to visualize.

    Other Parameters
    ----------------
    sample_distance: float, default: 0.2
        Sampling distance for volume rendering, normalized from 0.0 to 1.0.
        Lower values result in a higher quality rendering. High values improve
        the framerate.
    units: string, default: ''
        Units to display in the scale bar.

    Returns
    -------
    viewer : ipywidget
        Display by placing at the end of a Jupyter or Colab cell. Query or set
        properties on the object to change the visualization.
    """
    viewer = Viewer(data=data, **kwargs)

    return viewer



from enum import Enum
from importlib import import_module
import sys


class Env(Enum):
    JUPYTER_NOTEBOOK = 'notebook'
    JUPYTERLAB = 'lab'
    JUPYTERLITE = 'lite'
    SAGEMAKER = 'sagemaker'
    HYPHA = 'hypha'
    COLAB = 'colab'


def find_env():
    try:
        from google.colab import files
        return Env.COLAB
    except:
        try:
            from IPython import get_ipython
            parent_header = get_ipython().parent_header
            username = parent_header['header']['username']
            if username == '':
                return Env.JUPYTERLAB
            elif username == 'username':
                return Env.JUPYTER_NOTEBOOK
            else:
                return Env.SAGEMAKER
        except AttributeError:
            try:
                import js
                return Env.JUPYTERLITE
            except ImportError:
                return Env.HYPHA


ENVIRONMENT = find_env()

if ENVIRONMENT is not Env.JUPYTERLITE:
    if ENVIRONMENT is not Env.COLAB:
        if ENVIRONMENT is Env.JUPYTER_NOTEBOOK and sys.version_info.minor > 7:
            try:
                import imjoy_jupyter_extension
            except:
                raise RuntimeError('imjoy-jupyter-extension is required. `pip install itkwidgets[notebook]` and refresh page.')
        else:
            try:
                import_module("imjoy-jupyterlab-extension")
            except:
                if ENVIRONMENT is Env.JUPYTERLITE:
                    raise RuntimeError('imjoy-jupyterlab-extension is required. Install the package and refresh page.')
                elif sys.version_info.minor > 7:
                    raise RuntimeError('imjoy-jupyterlab-extension is required. `pip install itkwidgets[lab]` and refresh page.')

    try:
        import imjoy_elfinder
    except:
        if ENVIRONMENT is Env.JUPYTERLITE:
            raise RuntimeError('imjoy-elfinder is required. Install the package and refresh page.')
        elif sys.version_info.minor > 7:
            raise RuntimeError('imjoy-elfinder is required. `pip install imjoy-elfinder` and refresh page.')



def deferred_methods():
    return [
        'setImageBlendMode',
        'setImageColorMap',
        'setImageColorRange',
        'setImageColorRangeBounds',
        'setImageComponentVisibility',
        'setImageGradientOpacity',
        'setImageGradientOpacityScale',
        'setImageInterpolationEnabled',
        'setImagePiecewiseFunctionGaussians',
        'setImageVolumeSampleDistance',
        'setLabelImageBlend',
        'setLabelImageLabelNames',
        'setLabelImageLookupTable',
        'setLabelImageWeights',
        'selectLayer',
        'setLayerVisibility',
        'setUnits',
        'setViewMode',
        'setXSlice',
        'setYSlice',
        'setZSlice',
    ]



ITK_VIEWER_SRC = (
    "https://bafybeid7em5cvblohvq65tbk7wckdl4kbfdwvyqqikfjbjo54wsft5vu4a.on.fleek.co/"
)
PYDATA_SPHINX_HREF = "https://cdn.jsdelivr.net/npm/itk-viewer-bootstrap-ui@0.17.0/dist/bootstrapUIMachineOptions.js.es.js"
MUI_HREF = "https://cdn.jsdelivr.net/npm/itk-viewer-material-ui@0.3.0/dist/materialUIMachineOptions.js.es.js"



from dataclasses import asdict

from typing import Dict 

import itkwasm
import numcodecs
from imjoy_rpc import api
import zarr

_numcodec_encoder = numcodecs.Blosc(cname='lz4', clevel=3)
_numcodec_config = _numcodec_encoder.get_config()

def encode_itkwasm_image(image):
    global _numcodec_encoder

    image_dict = asdict(image)

    image_data = image_dict['data']
    encoded_data = _numcodec_encoder.encode(image_data)
    image_dict['data'] = { 'buffer': encoded_data, 'config': _numcodec_config, 'nbytes': image_data.nbytes }

    image_direction = image_dict['direction']
    encoded_direction = _numcodec_encoder.encode(image_direction)
    image_dict['direction'] = { 'buffer': encoded_direction, 'config': _numcodec_config, 'nbytes': image_direction.nbytes }

    return image_dict

def encode_zarr_store(store):
    def getItem(key):
        return store[key]

    def setItem(key, value):
        store[key] = value

    def containsItem(key):
        return key in store

    return {
        "_rintf": True,
        "_rtype": 'zarr-store',
        "getItem": getItem,
        "setItem": setItem,
        "containsItem": containsItem,
    }

def register_itkwasm_imjoy_codecs():

    api.registerCodec({'name': 'itkwasm-image', 'type': itkwasm.Image, 'encoder': encode_itkwasm_image})
    api.registerCodec({'name': 'zarr-store', 'type': zarr.storage.BaseStore, 'encoder': encode_zarr_store})
 



import itkwasm

from packaging import version
HAVE_ITK = False
try:
    import itk
    if not hasattr(itk, '__version__') or version.parse(itk.__version__) < version.parse('5.3.0'):
      raise RuntimeError('itk 5.3 or newer is required. `pip install itk>=5.3.0`')
    HAVE_ITK = True
except ImportError:
    pass
    

if HAVE_ITK:
    def itk_group_spatial_object_to_wasm_point_set(point_set):
        point_set_dict = itk.dict_from_pointset(point_set)
        wasm_point_set = itkwasm.PointSet(**point_set_dict)
        return wasm_point_set

else:
    def itk_group_spatial_object_to_wasm_point_set(point_set):
        raise RuntimeError('itk 5.3 or newer is required. `pip install itk>=5.3.0`')



import itkwasm
import numpy as np
import zarr
from ngff_zarr import to_multiscales, to_ngff_zarr, to_ngff_image, itk_image_to_ngff_image, Methods

import dask
from .itk import HAVE_ITK, itk_group_spatial_object_to_wasm_point_set
from .pytorch import HAVE_TORCH
from .vtk import HAVE_VTK, vtk_image_to_ngff_image, vtk_polydata_to_vtkjs
from .xarray import HAVE_XARRAY, HAVE_MULTISCALE_SPATIAL_IMAGE, xarray_data_array_to_numpy, xarray_data_set_to_numpy
from ..render_types import RenderType

def _spatial_image_scale_factors(spatial_image, min_length):
    sizes = dict(spatial_image.sizes)
    scale_factors = []
    dims = spatial_image.dims
    previous = { d: 1 for d in { 'x', 'y', 'z' }.intersection(dims) }
    while (np.array(list(sizes.values())) > min_length).any():
        max_size = np.array(list(sizes.values())).max()
        to_skip = { d: sizes[d] <= max_size / 2 for d in previous.keys() }
        scale_factor = {}
        for dim in previous.keys():
            if to_skip[dim]:
                scale_factor[dim] = previous[dim]
                continue
            scale_factor[dim] = 2 * previous[dim]

            sizes[dim] = int(sizes[dim] / 2)
        previous = scale_factor
        scale_factors.append(scale_factor)

    return scale_factors

def _make_multiscale_store():
    # Todo: for very large images serialize to disk cache
    # -> create DirectoryStore in cache directory and return as the chunk_store
    store = zarr.storage.MemoryStore(dimension_separator='/')
    return store, None

def _get_viewer_image(image, label=False):
    # NGFF Zarr
    if isinstance(image, zarr.Group) and 'multiscales' in image.attrs:
        return image.store

    min_length = 64
    if label:
        method = Methods.DASK_IMAGE_NEAREST
    else:
        method = Methods.DASK_IMAGE_GAUSSIAN

    store, chunk_store = _make_multiscale_store()

    if HAVE_MULTISCALE_SPATIAL_IMAGE:
        from multiscale_spatial_image import MultiscaleSpatialImage
        if isinstance(image, MultiscaleSpatialImage):
            image.to_zarr(store, compute=True)
            return store

    if isinstance(image, itkwasm.Image):
        ngff_image = itk_image_to_ngff_image(image)
        multiscales = to_multiscales(ngff_image, method=method)
        to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
        return store

    if HAVE_ITK:
        import itk
        if isinstance(image, itk.Image):
            ngff_image = itk_image_to_ngff_image(image)
            multiscales = to_multiscales(ngff_image, method=method)
            to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
            return store

    if HAVE_VTK:
        import vtk
        if isinstance(image, vtk.vtkImageData):
            ngff_image = vtk_image_to_ngff_image(image)
            multiscales = to_multiscales(ngff_image, method=method)
            to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
            return store

    if isinstance(image, dask.array.core.Array):
        ngff_image = to_ngff_image(image)
        multiscales = to_multiscales(ngff_image, method=method)
        to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
        return store

    if isinstance(image, zarr.Array):
        ngff_image = to_ngff_image(image)
        multiscales = to_multiscales(ngff_image, method=method)
        to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
        return store

    if HAVE_TORCH:
        import torch
        if isinstance(image, torch.Tensor):
            ngff_image = to_ngff_image(image.numpy())
            multiscales = to_multiscales(ngff_image, method=method)
            to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
            return store

    # Todo: preserve dask Array, if present, check if dims are NGFF -> use dims, coords
    # Check if coords are uniform, if not, resample
    if HAVE_XARRAY:
        import xarray as xr
        if isinstance(image, xr.DataArray):
            # if HAVE_MULTISCALE_SPATIAL_IMAGE:
            #     from spatial_image import is_spatial_image
            #     if is_spatial_image(image):
            #         from multiscale_spatial_image import to_multiscale
            #         scale_factors = _spatial_image_scale_factors(image, min_length)
            #         multiscale = to_multiscale(image, scale_factors, method=method)
            #         return _make_multiscale_store(multiscale)

            return xarray_data_array_to_numpy(image)
        if isinstance(image, xr.Dataset):
            # da = image[next(iter(image.variables.keys()))]
            # if is_spatial_image(da):
            #     scale_factors = _spatial_image_scale_factors(da, min_length)
            #     multiscale = to_multiscale(da, scale_factors, method=method)
            #     return _make_multiscale_store(multiscale)
            return xarray_data_set_to_numpy(image)

    if isinstance(image, np.ndarray):
        ngff_image = to_ngff_image(image)
        multiscales = to_multiscales(ngff_image, method=method)
        to_ngff_zarr(store, multiscales, chunk_store=chunk_store)
        return store

    raise RuntimeError("Could not process the viewer image")


def _get_viewer_point_set(point_set):
    if HAVE_VTK:
        import vtk
        if isinstance(point_set, vtk.vtkPolyData):
            return vtk_polydata_to_vtkjs(point_set)
    if isinstance(point_set, dask.array.core.Array):
        return np.asarray(point_set)
    if HAVE_TORCH:
        import torch
        if isinstance(point_set, torch.Tensor):
            return point_set.numpy()
    if HAVE_XARRAY:
        import xarray as xr
        if isinstance(point_set, xr.DataArray):
            return xarray_data_array_to_numpy(point_set)
        if isinstance(point_set, xr.Dataset):
            return xarray_data_set_to_numpy(point_set)
    if HAVE_ITK:
        import itk
        if isinstance(point_set, itk.PointSet):
            return itk.array_from_vector_container(point_set.GetPoints())
    return point_set


def _detect_render_type(data, input_type) -> RenderType:
    if input_type == 'image' or input_type == 'label_image':
        return RenderType.IMAGE
    elif input_type == 'point_set':
        return RenderType.POINT_SET
    if isinstance(data, itkwasm.Image):
        return RenderType.IMAGE
    elif isinstance(data, itkwasm.PointSet):
        return RenderType.POINT_SET
    elif isinstance(data, (zarr.Array, zarr.Group)):
        # For now assume zarr.Group is an image
        # In the future, once NGFF supports point sets fully
        # We may need to do more introspection
        return RenderType.IMAGE
    elif isinstance(data, np.ndarray):
        if data.ndim == 2 and data.shape[1] < 4:
            return RenderType.POINT_SET
        else:
            return RenderType.IMAGE
    elif HAVE_ITK:
        import itk
        if isinstance(data, itk.Image):
            return RenderType.IMAGE
        elif isinstance(data, itk.PointSet):
            return RenderType.POINT_SET
    if HAVE_MULTISCALE_SPATIAL_IMAGE:
        from multiscale_spatial_image import MultiscaleSpatialImage
        if isinstance(data, MultiscaleSpatialImage):
            return RenderType.IMAGE
    if HAVE_VTK:
        import vtk
        if isinstance(data, vtk.vtkImageData):
            return RenderType.IMAGE
        elif isinstance(data, vtk.vtkPolyData):
            return RenderType.POINT_SET
    if isinstance(data, dask.array.core.Array):
        if data.ndim ==2 and data.shape[1] < 4:
            return RenderType.POINT_SET
        else:
            return RenderType.IMAGE
    if HAVE_TORCH:
        import torch
        if isinstance(data, torch.Tensor):
            if data.dim == 2 and data.shape[1] < 4:
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
    if HAVE_XARRAY:
        import xarray as xr
        if isinstance(data, xr.DataArray):
            if data.dims == 2 and data.shape[1] < 4:
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
        if isinstance(data, xr.Dataset):
            if data.dims == 2 and data.shape[1] < 4:
                return RenderType.POINT_SET
            else:
                return RenderType.IMAGE
